#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_isis_sr_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2019 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
test_isis_sr_te_topo1.py:

                         +---------+
                         |         |
                         |   RT1   |
                         | 1.1.1.1 |
                         |         |
                         +---------+
                              |eth-sw1
                              |
                              |
                              |
         +---------+          |          +---------+
         |         |          |          |         |
         |   RT2   |eth-sw1   |   eth-sw1|   RT3   |
         | 2.2.2.2 +----------+----------+ 3.3.3.3 |
         |         |     10.0.1.0/24     |         |
         +---------+                     +---------+
    eth-rt4-1|  |eth-rt4-2          eth-rt5-1|  |eth-rt5-2
             |  |                            |  |
  10.0.2.0/24|  |10.0.3.0/24      10.0.4.0/24|  |10.0.5.0/24
             |  |                            |  |
    eth-rt2-1|  |eth-rt2-2          eth-rt3-1|  |eth-rt3-2
         +---------+                     +---------+
         |         |                     |         |
         |   RT4   |     10.0.6.0/24     |   RT5   |
         | 4.4.4.4 +---------------------+ 5.5.5.5 |
         |         |eth-rt5       eth-rt4|         |
         +---------+                     +---------+
       eth-rt6|                                |eth-rt6
              |                                |
   10.0.7.0/24|                                |10.0.8.0/24
              |          +---------+           |
              |          |         |           |
              |          |   RT6   |           |
              +----------+ 6.6.6.6 +-----------+
                  eth-rt4|         |eth-rt5
                         +---------+
                              |eth-dst (.1)
                              |
                              |10.0.11.0/24
                              |
                              |eth-rt6 (.2)
                         +---------+
                         |         |
                         |   DST   |
                         | 9.9.9.2 |
                         |         |
                         +---------+

"""

import os
import sys
import pytest
import json
from functools import partial

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.

pytestmark = [pytest.mark.bgpd, pytest.mark.isisd, pytest.mark.pathd]


def build_topo(tgen):
    "Build function"

    #
    # Define FRR Routers
    #
    for router in ["rt1", "rt2", "rt3", "rt4", "rt5", "rt6", "dst"]:
        tgen.add_router(router)

    #
    # Define connections
    #
    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["rt1"], nodeif="eth-sw1")
    switch.add_link(tgen.gears["rt2"], nodeif="eth-sw1")
    switch.add_link(tgen.gears["rt3"], nodeif="eth-sw1")

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["rt2"], nodeif="eth-rt4-1")
    switch.add_link(tgen.gears["rt4"], nodeif="eth-rt2-1")

    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["rt2"], nodeif="eth-rt4-2")
    switch.add_link(tgen.gears["rt4"], nodeif="eth-rt2-2")

    switch = tgen.add_switch("s4")
    switch.add_link(tgen.gears["rt3"], nodeif="eth-rt5-1")
    switch.add_link(tgen.gears["rt5"], nodeif="eth-rt3-1")

    switch = tgen.add_switch("s5")
    switch.add_link(tgen.gears["rt3"], nodeif="eth-rt5-2")
    switch.add_link(tgen.gears["rt5"], nodeif="eth-rt3-2")

    switch = tgen.add_switch("s6")
    switch.add_link(tgen.gears["rt4"], nodeif="eth-rt5")
    switch.add_link(tgen.gears["rt5"], nodeif="eth-rt4")

    switch = tgen.add_switch("s7")
    switch.add_link(tgen.gears["rt4"], nodeif="eth-rt6")
    switch.add_link(tgen.gears["rt6"], nodeif="eth-rt4")

    switch = tgen.add_switch("s8")
    switch.add_link(tgen.gears["rt5"], nodeif="eth-rt6")
    switch.add_link(tgen.gears["rt6"], nodeif="eth-rt5")

    switch = tgen.add_switch("s9")
    switch.add_link(tgen.gears["rt6"], nodeif="eth-dst")
    switch.add_link(tgen.gears["dst"], nodeif="eth-rt6")


def setup_module(mod):
    "Sets up the pytest environment"

    tgen = Topogen(build_topo, mod.__name__)

    frrdir = tgen.config.get(tgen.CONFIG_SECTION, "frrdir")
    if not os.path.isfile(os.path.join(frrdir, "pathd")):
        pytest.skip("pathd daemon wasn't built")

    tgen.start_topology()

    router_list = tgen.routers()

    # For all registered routers, load the zebra configuration file
    for rname, router in router_list.items():
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_ISIS, os.path.join(CWD, "{}/isisd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_PATH, os.path.join(CWD, "{}/pathd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()

    # This function tears down the whole topology.
    tgen.stop_topology()


def setup_testcase(msg):
    logger.info(msg)
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    return tgen


def print_cmd_result(rname, command):
    print(get_topogen().gears[rname].vtysh_cmd(command, isjson=False))


def compare_json_test(router, command, reference, exact):
    output = router.vtysh_cmd(command, isjson=True)
    result = topotest.json_cmp(output, reference)

    # Note: topotest.json_cmp() just checks on inclusion of keys.
    # For exact matching also compare the other way around.
    if not result and exact:
        return topotest.json_cmp(reference, output)
    else:
        return result


def cmp_json_output(rname, command, reference, exact=False):
    "Compare router JSON output"

    logger.info('Comparing router "%s" "%s" output', rname, command)

    tgen = get_topogen()
    filename = "{}/{}/{}".format(CWD, rname, reference)
    expected = json.loads(open(filename).read())

    # Run test function until we get an result. Wait at most 60 seconds.
    test_func = partial(compare_json_test, tgen.gears[rname], command, expected, exact)
    _, diff = topotest.run_and_expect(test_func, None, count=120, wait=0.5)
    assertmsg = '"{}" JSON output mismatches the expected result'.format(rname)
    assert diff is None, assertmsg


def cmp_json_output_exact(rname, command, reference):
    return cmp_json_output(rname, command, reference, True)


def compare_json_test_inverted(router, command, reference, exact):
    "logically inverts result of compare_json_test"

    # None vs something else
    result = compare_json_test(router, command, reference, exact)
    if result is None:
        return "Some"
    return None


def cmp_json_output_doesnt(rname, command, reference):
    "Compare router JSON output, shouldn't include reference"

    logger.info('Comparing (anti) router "%s" "%s" output', rname, command)

    tgen = get_topogen()
    filename = "{}/{}/{}".format(CWD, rname, reference)
    expected = json.loads(open(filename).read())

    # Run test function until we get an result. Wait at most 60 seconds.
    test_func = partial(
        compare_json_test_inverted, tgen.gears[rname], command, expected, exact=False
    )
    _, diff = topotest.run_and_expect(test_func, None, count=120, wait=0.5)
    assertmsg = '"{}" JSON output mismatches the expected result'.format(rname)
    assert diff is None, assertmsg


def dump_json(v):
    if isinstance(v, (dict, list)):
        return "\t" + "\t".join(
            json.dumps(v, indent=4, separators=(",", ": ")).splitlines(True)
        )
    else:
        return "'{}'".format(v)


def add_candidate_path(rname, endpoint, pref, name, segment_list="default", color=1):
    get_topogen().net[rname].cmd(
        """ \
        vtysh -c "conf t" \
              -c "segment-routing" \
              -c "traffic-eng" \
              -c "policy color """
        + str(color)
        + " endpoint "
        + endpoint
        + """" \
              -c "candidate-path preference """
        + str(pref)
        + """ name """
        + name
        + """ explicit segment-list """
        + segment_list
        + '''"'''
    )


def delete_candidate_path(rname, endpoint, pref, color=1):
    get_topogen().net[rname].cmd(
        """ \
        vtysh -c "conf t" \
              -c "segment-routing" \
              -c "traffic-eng" \
              -c "policy color """
        + str(color)
        + " endpoint "
        + endpoint
        + """" \
              -c "no candidate-path preference """
        + str(pref)
        + '''"'''
    )


def add_segment(rname, name, index, label):
    get_topogen().net[rname].cmd(
        """ \
        vtysh -c "conf t" \
              -c "segment-routing" \
              -c "traffic-eng" \
              -c "segment-list """
        + name
        + """" \
              -c "index """
        + str(index)
        + """ mpls label """
        + str(label)
        + '''"'''
    )


def delete_segment(rname, name, index):
    get_topogen().net[rname].cmd(
        """ \
        vtysh -c "conf t" \
              -c "segment-routing" \
              -c "traffic-eng" \
              -c "segment-list """
        + name
        + """" \
              -c "no index """
        + str(index)
        + '''"'''
    )


def create_sr_policy(rname, endpoint, bsid):
    get_topogen().net[rname].cmd(
        """ \
        vtysh -c "conf t" \
              -c "segment-routing" \
              -c "traffic-eng" \
              -c "policy color 1 endpoint """
        + endpoint
        + """" \
              -c "name default" \
              -c "binding-sid """
        + str(bsid)
        + '''"'''
    )


def delete_sr_policy(rname, endpoint):
    get_topogen().net[rname].cmd(
        """ \
        vtysh -c "conf t" \
              -c "segment-routing" \
              -c "traffic-eng" \
              -c "no policy color 1 endpoint """
        + endpoint
        + '''"'''
    )


def create_prefix_sid(rname, prefix, sid):
    get_topogen().net[rname].cmd(
        """ \
        vtysh -c "conf t" \
              -c "router isis 1" \
              -c "segment-routing prefix """
        + prefix
        + " index "
        + str(sid)
        + '''"'''
    )


def delete_prefix_sid(rname, prefix):
    get_topogen().net[rname].cmd(
        ''' \
        vtysh -c "conf t" \
              -c "router isis 1" \
              -c "no segment-routing prefix "'''
        + prefix
    )


def set_route_map_color(rname, color):
    get_topogen().net[rname].cmd(
        ''' \
        vtysh -c "conf t" \
              -c "route-map SET_SR_POLICY permit 10" \
              -c "set sr-te color  "'''
        + str(color)
    )


def router_bgp_shutdown_neighbor(rname, neighbor):
    get_topogen().net[rname].cmd(
        """ \
        vtysh -c "conf t" \
              -c "router bgp 1" \
              -c " neighbor """
        + neighbor
        + ' shutdown"'
    )


def router_bgp_no_shutdown_neighbor(rname, neighbor):
    get_topogen().net[rname].cmd(
        """ \
        vtysh -c "conf t" \
              -c "router bgp 1" \
              -c " no neighbor """
        + neighbor
        + ' shutdown"'
    )


def show_running_cfg(rname):
    output = (
        get_topogen()
        .net[rname]
        .cmd(
            """ \
        vtysh -c "show run" """
        )
    )
    logger.info(output)


#
# Step 1
#
# Checking the MPLS table using a single SR Policy and a single Candidate Path
#
def test_srte_init_step1():
    setup_testcase("Test (step 1): wait for IS-IS convergence / label distribution")

    for rname in ["rt1", "rt6"]:
        cmp_json_output(
            rname, "show mpls table json", "step1/show_mpls_table_without_candidate.ref"
        )


def test_srte_add_candidate_check_mpls_table_step1():
    setup_testcase("Test (step 1): check MPLS table regarding the added Candidate Path")

    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        add_candidate_path(rname, endpoint, 100, "default")
        cmp_json_output(
            rname, "show mpls table json", "step1/show_mpls_table_with_candidate.ref"
        )
        delete_candidate_path(rname, endpoint, 100)


def test_srte_reinstall_sr_policy_check_mpls_table_step1():
    setup_testcase(
        "Test (step 1): check MPLS table after the SR Policy was removed and reinstalled"
    )

    for rname, endpoint, bsid in [("rt1", "6.6.6.6", 1111), ("rt6", "1.1.1.1", 6666)]:
        add_candidate_path(rname, endpoint, 100, "default")
        delete_sr_policy(rname, endpoint)
        cmp_json_output(
            rname, "show mpls table json", "step1/show_mpls_table_without_candidate.ref"
        )
        create_sr_policy(rname, endpoint, bsid)
        add_candidate_path(rname, endpoint, 100, "default")
        cmp_json_output(
            rname, "show mpls table json", "step1/show_mpls_table_with_candidate.ref"
        )
        delete_candidate_path(rname, endpoint, 100)


#
# Step 2
#
# Checking pathd operational data using a single SR Policy and a single Candidate Path
#
def test_srte_bare_policy_step2():
    setup_testcase("Test (step 2): bare SR Policy should not be operational")

    for rname in ["rt1", "rt6"]:
        cmp_json_output_exact(
            rname,
            "show yang operational-data /frr-pathd:pathd pathd",
            "step2/show_operational_data.ref",
        )


def test_srte_add_candidate_check_operational_data_step2():
    setup_testcase(
        "Test (step 2): add single Candidate Path, SR Policy should be operational"
    )

    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        add_candidate_path(rname, endpoint, 100, "default")
        cmp_json_output(
            rname,
            "show yang operational-data /frr-pathd:pathd pathd",
            "step2/show_operational_data_with_candidate.ref",
        )


def test_srte_config_remove_candidate_check_operational_data_step2():
    setup_testcase(
        "Test (step 2): remove single Candidate Path, SR Policy should not be operational anymore"
    )

    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        delete_candidate_path(rname, endpoint, 100)
        cmp_json_output_exact(
            rname,
            "show yang operational-data /frr-pathd:pathd pathd",
            "step2/show_operational_data.ref",
        )


#
# Step 3
#
# Testing the Candidate Path selection
#
def test_srte_add_two_candidates_step3():
    setup_testcase("Test (step 3): second Candidate Path has higher Priority")

    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        for pref, cand_name in [("100", "first"), ("200", "second")]:
            add_candidate_path(rname, endpoint, pref, cand_name)
        cmp_json_output(
            rname,
            "show yang operational-data /frr-pathd:pathd pathd",
            "step3/show_operational_data_with_two_candidates.ref",
        )

    # cleanup
    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        for pref in ["100", "200"]:
            delete_candidate_path(rname, endpoint, pref)


def test_srte_add_two_candidates_with_reverse_priority_step3():
    setup_testcase("Test (step 3): second Candidate Path has lower Priority")

    # Use reversed priorities here
    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        for pref, cand_name in [("200", "first"), ("100", "second")]:
            add_candidate_path(rname, endpoint, pref, cand_name)
        cmp_json_output(
            rname,
            "show yang operational-data /frr-pathd:pathd pathd",
            "step3/show_operational_data_with_two_candidates.ref",
        )

    # cleanup
    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        for pref in ["100", "200"]:
            delete_candidate_path(rname, endpoint, pref)


def test_srte_remove_best_candidate_step3():
    setup_testcase("Test (step 3): delete the Candidate Path with higher priority")

    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        for pref, cand_name in [("100", "first"), ("200", "second")]:
            add_candidate_path(rname, endpoint, pref, cand_name)

    # Delete candidate with higher priority
    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        delete_candidate_path(rname, endpoint, 200)

    # Candidate with lower priority should get active now
    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        cmp_json_output(
            rname,
            "show yang operational-data /frr-pathd:pathd pathd",
            "step3/show_operational_data_with_single_candidate.ref",
        )
        # cleanup
        delete_candidate_path(rname, endpoint, 100)


#
# Step 4
#
# Checking MPLS table with a single SR Policy and a Candidate Path with different Segment Lists and other modifications
#
def test_srte_change_segment_list_check_mpls_table_step4():
    setup_testcase("Test (step 4): check MPLS table for changed Segment List")

    for rname, endpoint in [("rt1", "6.6.6.6"), ("rt6", "1.1.1.1")]:
        add_candidate_path(rname, endpoint, 100, "default")
        # now change the segment list name
        add_candidate_path(rname, endpoint, 100, "default", "test")
        cmp_json_output(rname, "show mpls table json", "step4/show_mpls_table.ref")
        delete_candidate_path(rname, endpoint, 100)


def test_srte_segment_list_add_segment_check_mpls_table_step4():
    setup_testcase(
        "Test (step 4): check MPLS table for added (then changed and finally deleted) segment"
    )

    add_candidate_path("rt1", "6.6.6.6", 100, "default", "test")

    # first add a new segment
    add_segment("rt1", "test", 25, 16050)
    cmp_json_output(
        "rt1", "show mpls table json", "step4/show_mpls_table_add_segment.ref"
    )

    # ... then change it ...
    add_segment("rt1", "test", 25, 16030)
    cmp_json_output(
        "rt1", "show mpls table json", "step4/show_mpls_table_change_segment.ref"
    )

    # ... and finally delete it
    delete_segment("rt1", "test", 25)
    cmp_json_output("rt1", "show mpls table json", "step4/show_mpls_table.ref")
    delete_candidate_path("rt1", "6.6.6.6", 100)


def save_rt(routername, filename):
    save_filename = routername + "/" + filename
    tgen = get_topogen()
    router = tgen.gears[routername]

    config_output = router.vtysh_cmd("sh run")

    route_output_json = json.loads(router.vtysh_cmd("show ip route bgp json"))
    route_output = dump_json(route_output_json)

    f = open(save_filename, "w")
    f.write(config_output)
    f.write(route_output)
    f.close()


#
# Step 5
#
# Checking the nexthop using a single SR Policy and a Candidate Path with configured route-map
#
def test_srte_route_map_with_sr_policy_check_nextop_step5():
    setup_testcase(
        "Test (step 5): recursive nexthop learned through BGP neighbour should be aligned with SR Policy from route-map"
    )

    # (re-)build the SR Policy two times to ensure that reinstalling still works
    for _ in [1, 2]:
        cmp_json_output(
            "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_inactive_srte.ref"
        )

        delete_sr_policy("rt1", "6.6.6.6")
        cmp_json_output(
            "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_inactive_srte.ref"
        )

        create_sr_policy("rt1", "6.6.6.6", 1111)
        cmp_json_output(
            "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_inactive_srte.ref"
        )

        add_candidate_path("rt1", "6.6.6.6", 100, "default")
        cmp_json_output(
            "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_active_srte.ref"
        )

        delete_candidate_path("rt1", "6.6.6.6", 100)


def test_srte_route_map_sr_policy_vs_route_order_step5():
    setup_testcase(
        "Test (step 5): Config policy first, add route after and check route validity"
    )

    #
    # BGP route and route-map are already configured.
    # route-map sets color 1 on BGP routes

    # Developer: to force pause here
    # tgen = get_topogen()
    # tgen.mininet_cli()

    #
    # Configure policy/path
    #
    add_candidate_path("rt1", "6.6.6.6", 100, "default", "default", 1)

    #
    # Route should be valid
    #
    logger.info(
        "BGP route and route-map are already configured, SR candidate path added after. Route should be valid"
    )
    cmp_json_output(
        "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_active_srte.ref"
    )

    #
    # shutdown/no-shutdown on BGP neighbor to delete/re-add BGP route
    #
    router_bgp_shutdown_neighbor("rt1", "6.6.6.6")
    router_bgp_no_shutdown_neighbor("rt1", "6.6.6.6")

    #
    # Route should be valid (but isn't)
    #
    logger.info(
        "After shutdown + no-shutdown neighbor. Route should be valid, but isn't"
    )
    cmp_json_output(
        "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_active_srte.ref"
    )

    #
    # delete and re-add policy/path
    #
    delete_candidate_path("rt1", "6.6.6.6", 100, 1)
    add_candidate_path("rt1", "6.6.6.6", 100, "default", "default", 1)

    #
    # Route should be valid
    #
    logger.info("After re-add candidate path. Route should be valid")
    cmp_json_output(
        "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_active_srte.ref"
    )

    # Developer: to force pause here
    # tgen = get_topogen()
    # tgen.mininet_cli()

    # clean up
    delete_candidate_path("rt1", "6.6.6.6", 100, 2)


def test_srte_route_map_sr_policy_vs_routemap_order_step5():
    setup_testcase(
        "Test (step 5): Config policy first, set route-map after and check route validity"
    )

    #
    # BGP route and route-map are already configured.
    # route-map sets color 1 on BGP routes
    #

    #
    # Configure policy/path
    #
    add_candidate_path("rt1", "6.6.6.6", 100, "default", "default", 1)

    #
    # Route should be valid
    #
    logger.info("After add candidate path. Route should be valid")
    cmp_json_output(
        "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_active_srte.ref"
    )

    # Developer: to force pause here
    # tgen = get_topogen()
    # tgen.mininet_cli()

    #
    # change route-map color to someting else and back again
    #
    set_route_map_color("rt1", 2)
    logger.info("route-map color was set to 2")
    # show_running_cfg("rt1")
    # 220625 nexthop no longer becomes empty. Colored routes without
    # matching SR policies now fall back to their non-colored equivalent
    # nexthops. So the route to 9.9.9.9/32 will now be valid, but with
    # different nexthop values.
    logger.info("now route table will lose policy-mapped route")
    cmp_json_output_doesnt(
        "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_active_srte.ref"
    )
    set_route_map_color("rt1", 1)
    logger.info("route-map color was set to 1")
    # show_running_cfg("rt1")

    #
    # Route should be valid (but isn't)
    #
    logger.info("After change route-map color. Route should be valid, but isn't")
    cmp_json_output(
        "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_active_srte.ref"
    )

    #
    # delete and re-add policy/path
    #
    delete_candidate_path("rt1", "6.6.6.6", 100, 1)
    add_candidate_path("rt1", "6.6.6.6", 100, "default", "default", 1)

    #
    # Route should be valid
    #
    logger.info("After delete/re-add candidate path. Route should be valid")
    cmp_json_output(
        "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_active_srte.ref"
    )

    # Developer: force pause
    # tgen = get_topogen()
    # tgen.mininet_cli()

    # clean up
    delete_candidate_path("rt1", "6.6.6.6", 100, 2)


def test_srte_route_map_with_sr_policy_reinstall_prefix_sid_check_nextop_step5():
    setup_testcase(
        "Test (step 5): remove and re-install prefix SID on fist path element and check SR Policy activity"
    )

    # first add a candidate path so the SR Policy is active
    add_candidate_path("rt1", "6.6.6.6", 100, "default")
    cmp_json_output(
        "rt1",
        "show yang operational-data /frr-pathd:pathd pathd",
        "step5/show_operational_data_active.ref",
    )

    # delete prefix SID from first element of the configured path and check
    # if the SR Policy is inactive since the label can't be resolved anymore
    delete_prefix_sid("rt5", "5.5.5.5/32")
    cmp_json_output(
        "rt1",
        "show yang operational-data /frr-pathd:pathd pathd",
        "step5/show_operational_data_inactive.ref",
    )
    cmp_json_output(
        "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_inactive_srte.ref"
    )

    # re-create the prefix SID and check if the SR Policy is active
    create_prefix_sid("rt5", "5.5.5.5/32", 50)
    cmp_json_output(
        "rt1",
        "show yang operational-data /frr-pathd:pathd pathd",
        "step5/show_operational_data_active.ref",
    )
    cmp_json_output(
        "rt1", "show ip route bgp json", "step5/show_ip_route_bgp_active_srte.ref"
    )


# Memory leak test template
def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC
#
# Copyright 2023 6WIND S.A.
# Authored by Farid Mihoub <farid.mihoub@6wind.com>
#

"""
bgp_extcomm_list-delete.py:

Test the following commands:
route-map test permit 10
  set extended-comm-list <arg> delete
"""

import functools
import json
import os
import pytest
import re
import sys

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib import topotest


pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    for routern in range(1, 3):
        tgen.add_router("r{}".format(routern))

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_convergence():
    tgen = get_topogen()
    r2 = tgen.gears["r2"]

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    def _bgp_converge():
        output = json.loads(r2.vtysh_cmd("show ip bgp neighbor 192.168.255.1 json"))
        expected = {
            "192.168.255.1": {
                "bgpState": "Established",
                "addressFamilyInfo": {
                    "ipv4Unicast": {
                        "acceptedPrefixCounter": 4,
                    }
                },
            }
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_converge)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Can't converge initially"


def _set_extcomm_list(gear, ecom_t, ecom):
    "Set the extended community for deletion."
    cmd = [
        "con t\n",
        f"bgp extcommunity-list standard r1-{ecom_t} permit {ecom_t} {ecom}\n",
        f"route-map r1-in permit 10\n",
        f"set extended-comm-list r1-{ecom_t} delete\n",
    ]
    gear.vtysh_cmd("".join(cmd))


def _bgp_extcomm_list_del_check(gear, prefix, ecom):
    """
    Check the non-presense of the extended community for the given prefix.
    """
    # get the extended community list attribute for the given prefix
    output = json.loads(gear.vtysh_cmd(f"show ip bgp {prefix} json"))
    ecoms = output.get("paths", [])[0].get("extendedCommunity", {})
    ecoms = ecoms.get("string")

    # ecoms might be None at the first time
    if not ecoms:
        return False
    return re.search(ecom, ecoms) is None


def test_rt_extcomm_list_delete():
    tgen = get_topogen()
    r2 = tgen.gears["r2"]

    # set the extended community for deletion
    _set_extcomm_list(r2, "rt", "1.1.1.1:1")

    # check for the deletion of the extended community
    test_func = functools.partial(
        _bgp_extcomm_list_del_check, r2, "10.10.10.1/32", r"1.1.1.1:1"
    )
    _, result = topotest.run_and_expect(test_func, True, count=60, wait=0.5)
    assert result, "RT extended community 1.1.1.1:1 was not stripped."


def test_soo_extcomm_list_delete():
    tgen = get_topogen()
    r2 = tgen.gears["r2"]

    # set the extended community for deletion
    _set_extcomm_list(r2, "soo", "2.2.2.2:2")

    # check for the deletion of the extended community
    test_func = functools.partial(
        _bgp_extcomm_list_del_check, r2, "10.10.10.2/32", r"2.2.2.2:2"
    )
    _, result = topotest.run_and_expect(test_func, True, count=60, wait=0.5)
    assert result, "SoO extended community 2.2.2.2:2 was not stripped."


def test_nt_extcomm_list_delete():
    tgen = get_topogen()
    r2 = tgen.gears["r2"]

    # set the extended community for deletion
    _set_extcomm_list(r2, "nt", "3.3.3.3:0")

    # check for the deletion of the extended community
    test_func = functools.partial(
        _bgp_extcomm_list_del_check, r2, "10.10.10.3/32", r"3.3.3.3"
    )
    _, result = topotest.run_and_expect(test_func, True, count=60, wait=0.5)
    assert result, "NT extended community 3.3.3.3:0 was not stripped."


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2022 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

"""
Test if BGP SoO per neighbor works correctly. Routes having SoO
extended community MUST be rejected if the neighbor is configured
with soo (neighbor soo).
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.common_config import step

pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    tgen.add_router("cpe1")
    tgen.add_router("cpe2")
    tgen.add_router("pe1")
    tgen.add_router("pe2")

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["cpe1"])
    switch.add_link(tgen.gears["pe1"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["pe1"])
    switch.add_link(tgen.gears["pe2"])

    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["pe2"])
    switch.add_link(tgen.gears["cpe2"])

    switch = tgen.add_switch("s4")
    switch.add_link(tgen.gears["cpe2"])
    switch.add_link(tgen.gears["cpe1"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    pe1 = tgen.gears["pe1"]
    pe2 = tgen.gears["pe2"]

    pe1.run("ip link add RED type vrf table 1001")
    pe1.run("ip link set up dev RED")
    pe2.run("ip link add RED type vrf table 1001")
    pe2.run("ip link set up dev RED")
    pe1.run("ip link set pe1-eth0 master RED")
    pe2.run("ip link set pe2-eth1 master RED")

    pe1.run("sysctl -w net.ipv4.ip_forward=1")
    pe2.run("sysctl -w net.ipv4.ip_forward=1")
    pe1.run("sysctl -w net.mpls.conf.pe1-eth0.input=1")
    pe2.run("sysctl -w net.mpls.conf.pe2-eth1.input=1")

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_OSPF, os.path.join(CWD, "{}/ospfd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_LDP, os.path.join(CWD, "{}/ldpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_soo():
    tgen = get_topogen()

    pe2 = tgen.gears["pe2"]

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    def _bgp_soo_unconfigured():
        output = json.loads(
            pe2.vtysh_cmd(
                "show bgp vrf RED ipv4 unicast neighbors 192.168.2.1 advertised-routes json"
            )
        )
        expected = {"advertisedRoutes": {"172.16.255.1/32": {"path": "65001"}}}
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_soo_unconfigured)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "Failed to see BGP convergence in pe2"

    step("Configure SoO (65000:1) for PE2 -- CPE2 session")
    pe2.vtysh_cmd(
        """
    configure terminal
    router bgp 65001 vrf RED
     address-family ipv4 unicast
      neighbor 192.168.2.1 soo 65000:1
    """
    )

    def _bgp_soo_configured():
        output = json.loads(
            pe2.vtysh_cmd(
                "show bgp vrf RED ipv4 unicast neighbors 192.168.2.1 advertised-routes json"
            )
        )
        expected = {"advertisedRoutes": {"172.16.255.1/32": None}}
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_soo_configured)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "SoO filtering does not work from pe2"

    step("Configure SoO (65000:2) for PE2 -- CPE2 session")
    pe2.vtysh_cmd(
        """
    configure terminal
    router bgp 65001 vrf RED
     address-family ipv4 unicast
      neighbor 192.168.2.1 soo 65000:2
    """
    )

    test_func = functools.partial(_bgp_soo_unconfigured)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "SoO filtering does not work from pe2"

    step("Unconfigure SoO for PE2 -- CPE2 session")
    pe2.vtysh_cmd(
        """
    configure terminal
    router bgp 65001 vrf RED
     address-family ipv4 unicast
      no neighbor 192.168.2.1 soo
    """
    )

    test_func = functools.partial(_bgp_soo_unconfigured)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.5)
    assert result is None, "SoO filtering does not work from pe2"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: NONE
# written 2016 by David Lamparter, placed in Public Domain.
import sys, markdown

template = """<html><head><meta charset="UTF-8"><style type="text/css">
body { max-width: 45em; margin: auto; margin-top: 2em; margin-bottom: 2em;
    font-family:Fira Sans,sans-serif; text-align: justify;
    counter-reset: ch2; }
pre, code { font-family:Fira Mono,monospace; }
pre > code { display: block; padding:0.5em; border:1px solid black;
    background-color:#eee; color:#000; }
h2:before { content: counter(ch2) ". "; counter-increment: ch2; }
h2 { clear: both; margin-top: 3em; text-decoration: underline; counter-reset: ch3; }
h3:before { content: counter(ch2) "." counter(ch3) ". "; counter-increment: ch3; }
h3 { clear: both; margin-top: 2em; font-weight: normal; font-style: italic; }
h4 { font-weight: normal; font-style: italic; }
img[alt~="float-right"] { float:right; margin-left:2em; margin-bottom:2em; }
</style></head><body>
%s
</body></html>
"""

md = markdown.Markdown(extensions=["extra", "toc"])

for fn in sys.argv[1:]:
    with open(fn, "r") as ifd:
        with open("%s.html" % (fn), "w") as ofd:
            ofd.write(
                (template % (md.convert(ifd.read().decode("UTF-8")))).encode("UTF-8")
            )



#!/usr/bin/env python
# SPDX-License-Identifier: ISC
#
# Copyright 2023 6WIND S.A.
# Authored by Dmytro Shytyi <dmytro.shytyi@6wind.com>
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted, provided
# that the above copyright notice and this permission notice appear
# in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NETDEF DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NETDEF BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
# OF THIS SOFTWARE.
#

import os
import sys
import json
import functools
import pytest

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger
from lib.common_config import required_linux_kernel_version
from lib.checkping import check_ping


def build_topo(tgen):
    r"""
     CE1     CE3      CE5
    (eth0)  (eth0)   (eth0)
      :2      :2      :2
       |       |       |
   192.168.1.0 |       |
      /24      |       |
      2001:   2001:   2001:
     1::/64  3::/64  5::/64
       |       |       |
      :1      :1      :1
   +-(eth1)--(eth2)---(eth3)-+
   |     \   /          |    |
   |    (vrf10)     (vrf20)  |
   |             R1          |
   +----------(eth0)---------+
                :1
                |
            2001::/64
                |
                :2
              (eth0)
    +----------(eth0)--------------+
    |            R2                |
    |   (vrf10)       (vrf20)      |
    |     /           /     \      |
    +-(eth1)-----(eth2)-----(eth3)-+
        :1         :1          :1
         |          |           |
      +------+   +------+   +------+
     /  2001: \ /  2001: \ /  2001: \
    /  2::/64  \ 4::/64  / \ 6::/64 /
   /192.168.2.0|        /   \      /
   \     /24  / \       |   |      |
      +------+   +------+   +------+
         |           |          |
        :2          :2         :2
      (eth0)      (eth0)      (eth0)
        CE2         CE4         CE6
    """

    tgen.add_router("r1")
    tgen.add_router("r2")
    tgen.add_router("ce1")
    tgen.add_router("ce2")
    tgen.add_router("ce3")
    tgen.add_router("ce4")
    tgen.add_router("ce5")
    tgen.add_router("ce6")

    tgen.add_link(tgen.gears["r1"], tgen.gears["r2"], "eth0", "eth0")
    tgen.add_link(tgen.gears["ce1"], tgen.gears["r1"], "eth0", "eth1")
    tgen.add_link(tgen.gears["ce2"], tgen.gears["r2"], "eth0", "eth1")
    tgen.add_link(tgen.gears["ce3"], tgen.gears["r1"], "eth0", "eth2")
    tgen.add_link(tgen.gears["ce4"], tgen.gears["r2"], "eth0", "eth2")
    tgen.add_link(tgen.gears["ce5"], tgen.gears["r1"], "eth0", "eth3")
    tgen.add_link(tgen.gears["ce6"], tgen.gears["r2"], "eth0", "eth3")


def setup_module(mod):
    result = required_linux_kernel_version("5.11")
    if result is not True:
        pytest.skip("Kernel requirements are not met")

    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()
    router_list = tgen.routers()
    for _, (rname, router) in enumerate(tgen.routers().items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.gears["r1"].run("modprobe vrf")
    tgen.gears["r1"].run("ip link add vrf10 type vrf table 10")
    tgen.gears["r1"].run("ip link set vrf10 up")
    tgen.gears["r1"].run("ip link add vrf20 type vrf table 20")
    tgen.gears["r1"].run("ip link set vrf20 up")
    tgen.gears["r1"].run("ip link set eth1 master vrf10")
    tgen.gears["r1"].run("ip link set eth2 master vrf10")
    tgen.gears["r1"].run("ip link set eth3 master vrf20")
    tgen.gears["r1"].run("sysctl net.vrf.strict_mode=1")
    tgen.gears["r1"].run("sysctl net.ipv4.conf.default.rp_filter=0")
    tgen.gears["r1"].run("sysctl net.ipv4.conf.all.rp_filter=0")
    tgen.gears["r1"].run("sysctl net.ipv4.conf.lo.rp_filter=0")
    tgen.gears["r1"].run("sysctl net.ipv4.conf.eth0.rp_filter=0")
    tgen.gears["r1"].run("sysctl net.ipv4.conf.eth1.rp_filter=0")
    tgen.gears["r1"].run("sysctl net.ipv4.conf.vrf10.rp_filter=0")

    tgen.gears["r2"].run("modprobe vrf")
    tgen.gears["r2"].run("ip link add vrf10 type vrf table 10")
    tgen.gears["r2"].run("ip link set vrf10 up")
    tgen.gears["r2"].run("ip link add vrf20 type vrf table 20")
    tgen.gears["r2"].run("ip link set vrf20 up")
    tgen.gears["r2"].run("ip link set eth1 master vrf10")
    tgen.gears["r2"].run("ip link set eth2 master vrf20")
    tgen.gears["r2"].run("ip link set eth3 master vrf20")
    tgen.gears["r2"].run("sysctl net.vrf.strict_mode=1")
    tgen.gears["r2"].run("sysctl net.ipv4.conf.default.rp_filter=0")
    tgen.gears["r2"].run("sysctl net.ipv4.conf.all.rp_filter=0")
    tgen.gears["r2"].run("sysctl net.ipv4.conf.lo.rp_filter=0")
    tgen.gears["r2"].run("sysctl net.ipv4.conf.eth0.rp_filter=0")
    tgen.gears["r2"].run("sysctl net.ipv4.conf.eth1.rp_filter=0")
    tgen.gears["r2"].run("sysctl net.ipv4.conf.vrf10.rp_filter=0")
    tgen.start_router()

    # FOR DEVELOPER:
    # If you want to stop some specific line and start interactive shell,
    # please use tgen.mininet_cli() to start it.
    # Example:
    # tgen=get_topogen()
    # tgen.mininet_cli()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def open_json_file(filename):
    try:
        with open(filename, "r") as f:
            return json.load(f)
    except IOError:
        assert False, "Could not read file {}".format(filename)


def check_rib(name, cmd, expected_file):
    def _check(name, cmd, expected_file):
        logger.info("polling")
        tgen = get_topogen()
        router = tgen.gears[name]
        output = json.loads(router.vtysh_cmd(cmd))
        expected = open_json_file("{}/{}".format(CWD, expected_file))
        return topotest.json_cmp(output, expected)

    logger.info('[+] check {} "{}" {}'.format(name, cmd, expected_file))
    tgen = get_topogen()
    func = functools.partial(_check, name, cmd, expected_file)
    _, result = topotest.run_and_expect(func, None, count=10, wait=0.5)
    assert result is None, "Failed"


def test_rib():
    check_rib("r1", "show bgp ipv6 vpn json", "r1/vpnv6_rib.json")
    check_rib("r2", "show bgp ipv6 vpn json", "r2/vpnv6_rib.json")
    check_rib("r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_rib.json")
    check_rib("r1", "show ipv6 route vrf vrf20 json", "r1/vrf20_rib.json")
    check_rib("r2", "show ipv6 route vrf vrf10 json", "r2/vrf10_rib.json")
    check_rib("r2", "show ipv6 route vrf vrf20 json", "r2/vrf20_rib.json")
    check_rib("ce1", "show ipv6 route json", "ce1/ipv6_rib.json")
    check_rib("ce2", "show ipv6 route json", "ce2/ipv6_rib.json")
    check_rib("ce3", "show ipv6 route json", "ce3/ipv6_rib.json")
    check_rib("ce4", "show ipv6 route json", "ce4/ipv6_rib.json")
    check_rib("ce5", "show ipv6 route json", "ce5/ipv6_rib.json")
    check_rib("ce6", "show ipv6 route json", "ce6/ipv6_rib.json")


def test_ping():
    check_ping("ce1", "2001:2::2", True, 10, 0.5)
    check_ping("ce1", "2001:3::2", True, 10, 0.5)
    check_ping("ce1", "2001:4::2", False, 10, 0.5)
    check_ping("ce1", "2001:5::2", False, 10, 0.5)
    check_ping("ce1", "2001:6::2", False, 10, 0.5)
    check_ping("ce4", "2001:1::2", False, 10, 0.5)
    check_ping("ce4", "2001:2::2", False, 10, 0.5)
    check_ping("ce4", "2001:3::2", False, 10, 0.5)
    check_ping("ce4", "2001:5::2", True, 10, 0.5)
    check_ping("ce4", "2001:6::2", True, 10, 0.5)


def test_sid_per_afv6_auto():
    check_rib("r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_afv6_auto_sid_rib.json")
    check_ping("ce1", "2001:2::2", True, 10, 0.5)
    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          address-family ipv6 unicast
           no sid vpn export auto
        """
    )
    check_rib(
        "r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_afv6_auto_no_sid_rib.json"
    )
    check_ping("ce1", "2001:2::2", False, 10, 0.5)

    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          address-family ipv6 unicast
           sid vpn export auto
        """
    )
    check_rib("r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_afv6_auto_sid_rib.json")
    check_ping("ce1", "2001:2::2", True, 10, 0.5)

    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          address-family ipv6 unicast
           no sid vpn export auto
        """
    )
    check_rib(
        "r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_afv6_auto_no_sid_rib.json"
    )
    check_ping("ce1", "2001:2::2", False, 10, 0.5)


def test_sid_per_afv6_manual():
    check_rib(
        "r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_afv6_manual_no_sid_rib.json"
    )
    check_ping("ce1", "2001:2::2", False, 10, 0.5)

    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          address-family ipv6 unicast
           sid vpn export 8
        """
    )

    check_rib(
        "r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_afv6_manual_sid_rib.json"
    )
    check_ping("ce1", "2001:2::2", True, 10, 0.5)

    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          address-family ipv6 unicast
           no sid vpn export 8
        """
    )
    check_rib(
        "r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_afv6_manual_no_sid_rib.json"
    )
    check_ping("ce1", "2001:2::2", False, 10, 0.5)


def test_sid_per_afv4_auto():
    check_rib("r1", "show ip route vrf vrf10 json", "r1/vrf10_afv4_auto_sid_rib.json")
    check_ping("ce1", "192.168.2.2", True, 10, 0.5)
    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          address-family ipv4 unicast
           no sid vpn export auto
        """
    )

    check_rib(
        "r1", "show ip route vrf vrf10 json", "r1/vrf10_afv4_auto_no_sid_rib.json"
    )
    check_ping("ce1", "192.168.2.2", False, 10, 0.5)

    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          address-family ipv4 unicast
           sid vpn export auto
        """
    )

    check_rib("r1", "show ip route vrf vrf10 json", "r1/vrf10_afv4_auto_sid_rib.json")
    check_ping("ce1", "192.168.2.2", True, 10, 0.5)

    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          address-family ipv4 unicast
           no sid vpn export auto
        """
    )
    check_rib(
        "r1", "show ip route vrf vrf10 json", "r1/vrf10_afv4_auto_no_sid_rib.json"
    )
    check_ping("ce1", "192.168.2.2", False, 10, 0.5)


def test_sid_per_afv4_manual():
    check_rib(
        "r1", "show ip route vrf vrf10 json", "r1/vrf10_afv4_manual_no_sid_rib.json"
    )
    check_ping("ce1", "192.168.2.2", False, 10, 0.5)
    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          address-family ipv4 unicast
           sid vpn export 8
        """
    )

    check_rib("r1", "show ip route vrf vrf10 json", "r1/vrf10_afv4_manual_sid_rib.json")
    check_ping("ce1", "192.168.2.2", True, 10, 0.5)

    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          address-family ipv4 unicast
           no sid vpn export 8
        """
    )
    check_ping("ce1", "192.168.2.2", False, 10, 0.5)
    check_rib(
        "r1", "show ip route vrf vrf10 json", "r1/vrf10_afv4_manual_no_sid_rib.json"
    )


def test_sid_per_vrf_auto():
    check_rib(
        "r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_pervrf_auto_no_sid_rib.json"
    )
    check_ping("ce1", "2001:2::2", False, 10, 0.5)
    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          sid vpn per-vrf export auto
        """
    )

    check_rib(
        "r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_pervrf6_auto_sid_rib.json"
    )
    check_ping("ce1", "2001:2::2", True, 10, 0.5)
    check_rib(
        "r1", "show ip route vrf vrf10 json", "r1/vrf10_pervrf4_auto_sid_rib.json"
    )
    check_ping("ce1", "192.168.2.2", True, 10, 0.5)

    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
           no sid vpn per-vrf export auto
        """
    )

    check_rib(
        "r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_pervrf_auto_no_sid_rib.json"
    )
    check_ping("ce1", "2001:2::2", False, 10, 0.5)


def test_sid_per_vrf_manual():
    check_rib(
        "r1", "show ip route vrf vrf10 json", "r1/vrf10_pervrf_manual_no_sid_rib.json"
    )
    check_ping("ce1", "192.168.2.2", False, 10, 0.5)
    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
          sid vpn per-vrf export 8
        """
    )

    check_rib(
        "r1", "show ipv6 route vrf vrf10 json", "r1/vrf10_pervrf6_manual_sid_rib.json"
    )
    check_ping("ce1", "2001:2::2", True, 10, 0.5)
    check_rib(
        "r1", "show ip route vrf vrf10 json", "r1/vrf10_pervrf4_manual_sid_rib.json"
    )
    check_ping("ce1", "192.168.2.2", True, 10, 0.5)

    get_topogen().gears["r2"].vtysh_cmd(
        """
        configure terminal
         router bgp 2 vrf vrf10
           no sid vpn per-vrf export 8
        """
    )

    check_rib(
        "r1", "show ip route vrf vrf10 json", "r1/vrf10_pervrf_manual_no_sid_rib.json"
    )
    check_ping("ce1", "192.168.2.2", False, 10, 0.5)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



import subprocess
import sys
import shlex
import os
import re

os.environ["LC_ALL"] = "C"
os.environ["LANG"] = "C"
for k in list(os.environ.keys()):
    if k.startswith("LC_"):
        os.environ.pop(k)

if len(sys.argv) < 2:
    sys.stderr.write("start as format-test.py gcc-123.45 [-options ...]\n")
    sys.exit(1)

c_re = re.compile(r"//\s+(NO)?WARN")
expect = {}
lines = {}

with open("format-test.c", "r") as fd:
    for lno, line in enumerate(fd.readlines(), 1):
        lines[lno] = line.strip()
        m = c_re.search(line)
        if m is None:
            continue
        if m.group(1) is None:
            expect[lno] = "warn"
        else:
            expect[lno] = "nowarn"

cmd = shlex.split(
    "-Wall -Wextra -Wno-unused -fplugin=./frr-format.so -fno-diagnostics-show-caret -c -o format-test.o format-test.c"
)

gcc = subprocess.Popen(
    sys.argv[1:] + cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
)
sout, serr = gcc.communicate()
gcc.wait()

gcclines = serr.decode("UTF-8").splitlines()
line_re = re.compile(r"^format-test\.c:(\d+):(.*)$")
gcc_warns = {}

for line in gcclines:
    if line.find("In function") >= 0:
        continue
    m = line_re.match(line)
    if m is None:
        sys.stderr.write("cannot process GCC output: %s\n" % line)
        continue

    lno = int(m.group(1))
    gcc_warns.setdefault(lno, []).append(line)

for lno, val in expect.items():
    if val == "nowarn" and lno in gcc_warns:
        sys.stderr.write(
            "unexpected gcc warning on line %d:\n\t%s\n\t%s\n"
            % (lno, lines[lno], "\n\t".join(gcc_warns[lno]))
        )
    if val == "warn" and lno not in gcc_warns:
        sys.stderr.write(
            "expected warning on line %d but did not get one\n\t%s\n"
            % (lno, lines[lno])
        )

leftover = set(gcc_warns.keys()) - set(expect.keys())
for lno in sorted(leftover):
    sys.stderr.write(
        "unmarked gcc warning on line %d:\n\t%s\n\t%s\n"
        % (lno, lines[lno], "\n\t".join(gcc_warns[lno]))
    )
